
package uk.co.bedroomcoders.jfxTest;

import javafx.application.Application;
import javafx.application.Platform;
import javafx.scene.Group;
import javafx.scene.Node;
import javafx.stage.Stage;
import javafx.scene.Scene;

import javafx.scene.paint.Color;
import javafx.scene.paint.PhongMaterial;
import javafx.scene.shape.Box;
import javafx.scene.shape.Sphere;
import javafx.scene.shape.CullFace;
import javafx.scene.PointLight;
import javafx.scene.AmbientLight;

import javafx.scene.PerspectiveCamera;

import javafx.event.ActionEvent;
import javafx.event.EventHandler;

import javafx.util.Duration;
import javafx.animation.KeyFrame;
import javafx.animation.Timeline;

import javafx.scene.input.KeyEvent;
import javafx.scene.input.KeyCode;
import javafx.scene.input.MouseEvent;

import com.sun.javafx.geom.Vec3d;

public class Main extends Application {
    
    private static final double AXIS_LENGTH = 250.0;
	private static final Vec3d up = new Vec3d(0, -1, 0); 

	private double a=0; // everything rotates using multiples of this counter

    private final Group root = new Group();
    private final PerspectiveCamera cam = new PerspectiveCamera(true);

    private Pivot axis = new Pivot("axis");
    private Pivot camera = new Pivot("camera");
    private Pivot earth = new Pivot("earth");
    private Pivot light = new Pivot("light");
    private Pivot redLight = new Pivot("red light");
    private Pivot box = new Pivot("box");
    
    private double lastMouseX,lastMouseY,mouseX,mouseY;
    private double camXa = 10,camYa = 345;
    private Vec3d fwd = new Vec3d(), tmpV3 = new Vec3d();
    
    private boolean moving = false;
    private boolean shiftDown = false;
    
    private PivotGui gui;
	
	public static void main(String[] args) { launch(args); }
	    
    private void buildAxes(Pivot p) {
        final Box xAxis = new Box(AXIS_LENGTH, 4, 4);
        final Box yAxis = new Box(4, AXIS_LENGTH, 4);
        final Box zAxis = new Box(4, 4, AXIS_LENGTH);

        xAxis.setMaterial(PhongPhactory.fromColour(Color.RED));
        yAxis.setMaterial(PhongPhactory.fromColour(Color.GREEN));
        zAxis.setMaterial(PhongPhactory.fromColour(Color.BLUE));

        p.getChildren().addAll(xAxis, yAxis, zAxis);
        p.setVisible(true);
        root.getChildren().addAll(p);
    }

    @Override
    public void start(Stage stage) {
        Scene scene = new Scene(root, 640, 480, true);        
        scene.setFill(Color.GREY);
		root.getChildren().add(new AmbientLight(Color.color(0.1,0.1,0.1)));
		gui = new PivotGui(root);
		
		buildAxes(axis);
		Pivot p4 = new Pivot("NE axis",  AXIS_LENGTH/2, 0,  AXIS_LENGTH/2); buildAxes(p4);
		Pivot p2 = new Pivot("SE axis",  AXIS_LENGTH/2, 0, -AXIS_LENGTH/2); buildAxes(p2);
		Pivot p1 = new Pivot("SW axis", -AXIS_LENGTH/2, 0, -AXIS_LENGTH/2); buildAxes(p1);
		Pivot p3 = new Pivot("NW axis", -AXIS_LENGTH/2, 0,  AXIS_LENGTH/2); buildAxes(p3);

		Sphere s = new Sphere(100);
		s.setMaterial(PhongPhactory.fromImage("data/ball.jpg"));
		earth.getChildren().add(s);
		root.getChildren().add(earth);

		camera.setPosition(140,-230,-540);
		camera.getChildren().add(cam);
		
		cam.setNearClip(0.1);
		cam.setFarClip(2000.0);
		cam.setFieldOfView(35);
		scene.setCamera(cam);
		
		PointLight l = new PointLight(Color.WHITE);
		light.getChildren().add(l);
		light.setPosition(300,-200,-300);
		//root.getChildren().add(light);
		buildAxes(light); // adding the axis will add light pivot to scene
		
		PointLight rl = new PointLight(Color.DARKRED);
		redLight.getChildren().add(rl);
		redLight.setPosition(-300,-200,300);
		buildAxes(redLight);
		
		Box b=new Box(1200,600,1200);
		box.getChildren().add(b);
		box.setPosition(0,0,0);
		b.setCullFace(CullFace.FRONT);
		b.setMaterial(PhongPhactory.fromImage("data/crate.jpg"));
		root.getChildren().add(box);
		
		stage.setScene(scene);
        stage.show();

		Timeline t = new Timeline( new KeyFrame(Duration.millis(20), e->update()));
		t.setCycleCount(Timeline.INDEFINITE);
		t.play();

		scene.setOnKeyPressed(e->keyDown(e));
		scene.setOnKeyReleased(e->keyUp(e));
        scene.setOnMouseDragged(e->mouseDrag(e));
        scene.setOnMousePressed(e->mousePressed(e));
        scene.setOnMouseReleased(e->mouseReleased(e));
	}	
	
	public void mousePressed(MouseEvent me) {
		lastMouseX = me.getSceneX(); lastMouseY = me.getSceneY();
		mouseX = me.getSceneX(); mouseY = me.getSceneY();
		
		if (me.isSecondaryButtonDown()) moving = true;
		Node n = me.getPickResult().getIntersectedNode();
		
		if (n!=null && shiftDown) { 
			// navigate up the hiarchy until we hit a Pivot
			while(n.getClass()!=Pivot.class && n!=null) n = n.getParent();
			
			if (n!=null) {  // found a clicked Pivot
				System.out.println(n);
				if (n.getId().equals("box")) {
					gui.setVisible(false);
				} else {
					gui.update((Pivot)n);
				}
			}	
		}
	}
	
	public void mouseReleased(MouseEvent me) {
		if (!me.isSecondaryButtonDown()) moving = false;	
	}

	public void mouseDrag(MouseEvent me) {
		lastMouseX = mouseX; lastMouseY = mouseY;
		mouseX = me.getSceneX(); mouseY = me.getSceneY();
		
		camYa=bound(camYa+((mouseX-lastMouseX)/4.0));
		camXa=bound(camXa+((mouseY-lastMouseY)/4.0));
	}
	
	// not strictly necessary 
	public double bound(double val) {
		if (val<0) val+=360;
		if (val>360) val-=360;
		return val;
	}
	
	private final Pivot origin = new Pivot(0,0,0);
	
	public void update() {

		a=a+0.2;
		
		earth.setEularRotation(a,-a*4,-a);
		
		axis.setPosition(Math.cos(a/16)*(AXIS_LENGTH/2), Math.cos(a/12)*Math.sin(a/14)*(AXIS_LENGTH/4), Math.sin(a/18)*(AXIS_LENGTH/2));
		axis.lookAt(origin.getPosition(), up);
		
		camera.setEularRotation(camXa,-camYa,0);
		
		if (moving) {
			fwd.set(0,0,1.5);
			fwd=camera.rotateVector(fwd);
			tmpV3 = camera.getPosition();
			tmpV3.add(fwd);
			camera.setPosition(tmpV3);
		}
	}
	
	public void keyDown(KeyEvent ke) {
		if (ke.getCode()==KeyCode.SHIFT) shiftDown=true;
	}
	
	public void keyUp(KeyEvent ke) { 
		if (ke.getCode()==KeyCode.ESCAPE) Platform.exit(); 
		if (ke.getCode()==KeyCode.SHIFT) shiftDown=false;
	}
}
